@@ -84,13 +84,13 @@ def train_from_folder(
     generate = False,
     generate_interpolation = False,
     hamburger_res_layers = [32],
-    use_evonorm = False,
+    disc_output_size = 5,
     interpolation_num_steps = 100,
     save_frames = False,
     num_image_tiles = 8,
     trunc_psi = 0.75,
     aug_prob = 0.,
-    aug_types = ['cutout', 'color'],
+    aug_types = ['cutout'],
     dataset_aug_prob = 0.,
     multi_gpus = False,
     calculate_fid_every = None,
@@ -104,7 +104,7 @@ def train_from_folder(
         batch_size = batch_size,
         gradient_accumulate_every = gradient_accumulate_every,
         hamburger_res_layers = cast_list(hamburger_res_layers),
-        use_evonorm = use_evonorm,
+        disc_output_size = disc_output_size,
         image_size = image_size,
         fmap_max = fmap_max,
         transparent = transparent,
